from typing import Optional

from ddtrace import Tracer, tracer as global_tracer
from ddtrace.constants import ANALYTICS_SAMPLE_RATE_KEY
from ddtrace.ext import http as http_tags
from ddtrace.http import store_request_headers, store_response_headers
from ddtrace.propagation.http import HTTPPropagator
from ddtrace.settings import config
from starlette.datastructures import Headers
from starlette.requests import Request
from starlette.types import ASGIApp, Message, Receive, Scope, Send


class TraceMiddleware:
    def __init__(
        self,
        app: ASGIApp,
        *,
        tracer: Optional[Tracer] = None,
        service: str = "asgi",
        distributed_tracing: bool = True,
    ) -> None:
        self.app = app
        if tracer is None:
            tracer = global_tracer
        self.tracer = tracer
        self.service = service
        self._distributed_tracing = distributed_tracing

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if scope["type"] != "http":
            await self.app(scope, receive, send)
            return

        request = Request(scope=scope, receive=receive)

        if self._distributed_tracing:
            propagator = HTTPPropagator()
            context = propagator.extract(request.headers)
            if context.trace_id:
                self.tracer.context_provider.activate(context)

        resource = "%s %s" % (request.method, request.url.path)
        span = self.tracer.trace(
            name="asgi.request",
            service=self.service,
            resource=resource,
            span_type=http_tags.TYPE,
        )

        span.set_tag(
            ANALYTICS_SAMPLE_RATE_KEY,
            config.asgi.get_analytics_sample_rate(use_global_config=True),
        )
        span.set_tag(http_tags.METHOD, request.method)
        span.set_tag(http_tags.URL, str(request.url))
        if config.asgi.trace_query_string:
            span.set_tag(http_tags.QUERY_STRING, request.url.query)

        # NOTE: any request header set in the future will not be stored in the span.
        store_request_headers(request.headers, span, config.asgi)

        async def send_with_tracing(message: Message) -> None:
            span = self.tracer.current_span()

            if not span:
                # Unexpected.
                await send(message)
                return

            if message["type"] == "http.response.start":
                status_code: int = message["status"]
                response_headers = Headers(raw=message["headers"])
                store_response_headers(response_headers, span, config.asgi)
                span.set_tag(http_tags.STATUS_CODE, str(status_code))

            await send(message)

        try:
            await self.app(scope, receive, send_with_tracing)
        except BaseException as exc:
            span.set_traceback()
            raise exc from None
        finally:
            span.finish()



import re
import typing
from pathlib import Path

from setuptools import setup


def get_version(package: str) -> str:
    version = Path(package, "__version__.py").read_text()
    match = re.search("__version__ = ['\"]([^'\"]+)['\"]", version)
    assert match is not None
    return match.group(1)


def get_long_description() -> str:
    with open("README.md", encoding="utf8") as readme:
        with open("CHANGELOG.md", encoding="utf8") as changelog:
            return readme.read() + "\n\n" + changelog.read()


def get_packages(package: str) -> typing.List[str]:
    return [str(path.parent) for path in Path(package).glob("**/__init__.py")]


setup(
    name="ddtrace-asgi",
    version=get_version("ddtrace_asgi"),
    description="Unofficial ddtrace integration for ASGI apps and frameworks",
    long_description=get_long_description(),
    long_description_content_type="text/markdown",
    url="http://github.com/florimondmanca/ddtrace-asgi",
    author="Florimond Manca",
    author_email="florimond.manca@gmail.com",
    packages=get_packages("ddtrace_asgi"),
    include_package_data=True,
    package_data={"ddtrace_asgi": ["py.typed"]},
    zip_safe=False,
    install_requires=["ddtrace", "starlette==0.*"],
    python_requires=">=3.6",
    license="BSD",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
)
